diff --git a/src/backend/access/transam/xlog.c b/src/backend/access/transam/xlog.c
index 24165ab03e..3e74bec42d 100644
--- a/src/backend/access/transam/xlog.c
+++ b/src/backend/access/transam/xlog.c
@@ -3127,6 +3127,16 @@ XLogBackgroundFlush(void)
 		flexible = false;		/* ensure it all gets written */
 	}
 
+	/* if we have flushed that far, consider segment boundaries */
+	if (WriteRqst.Write <= LogwrtResult.Flush)
+	{
+		SpinLockAcquire(&XLogCtl->segtrack_lck);
+		if (XLogCtl->earliestSegBoundary != MaxXLogSegNo)
+			WriteRqst.Write = XLogCtl->earliestSegBoundaryEndPtr;
+		SpinLockRelease(&XLogCtl->segtrack_lck);
+		Assert(!flexible);
+	}
+
 	/*
 	 * If already known flushed, we're done. Just need to check if we are
 	 * holding an open file handle to a logfile that's no longer in use,
diff --git a/src/backend/postmaster/walwriter.c b/src/backend/postmaster/walwriter.c
index 6a1e16edc2..8ca0d8e616 100644
--- a/src/backend/postmaster/walwriter.c
+++ b/src/backend/postmaster/walwriter.c
@@ -253,7 +253,8 @@ WalWriterMain(void)
 		 * here to handle a race condition where WAL is flushed to disk prior
 		 * to registering the segment boundary.
 		 */
-		NotifySegmentsReadyForArchive(GetFlushRecPtr());
+		if (XLogArchivingActive())
+			NotifySegmentsReadyForArchive(GetFlushRecPtr());
 
 		/*
 		 * Do what we're here for; then, if XLogBackgroundFlush() found useful
